#!BPY

"""
Name: 'The Movies Animation (.anm) Rev 6.1'
Blender: 246
Group: 'Import'
Tooltip: 'Import from Lionhead Studios The Movies Animation file format. (.anm)'
"""

__author__ = ["Glen Rickey"]
__url__ = ("Director's Cut Modding Foundry","http://www.dcmodding.com")
__version__ = "1.00 05-12-2009"
__bpydoc__ = """\

Anm Importer  (01142019)

This script imports an anm file from Lionhead Studio Limited's The Movies
into blender for editing. It is not supported by either Lionhead or
Activision.

Known Problems:

Expects to be imported into an existing TM mesh that has been imported with the proper armature bone structure and/or correct group_names.


"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C) Mark S Andrews, Nick Hudson, Glen Rickey
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------


# Import Modules
import Blender
from Blender import Image, Material, Texture, Window, Armature
from Blender import Mesh, NMesh, Group
from Blender import Ipo
from Blender.Mathutils import *
import struct, os, math
import tmConst,bonelist
import time,pdb
AH_CURRENT_VERSION = 12
compression = []
origins = []
scn = Blender.Scene.GetCurrent()
def UncompressU16(val):
    return (val/32767.0 - 1)
    #return (val/32767.0 )



#***********************************************
# helper classes & functions for files
#***********************************************

def read_item(struct_type, stream):
    value = struct.unpack(struct_type,
        stream.read(struct.calcsize(struct_type)))
    if len(value) == 1:
        value = value[0]
    return value

def read_nts(stream, length=32):
    s = stream.read(length)
    return s.rstrip("\x00")

def pad(stream):
    align = stream.tell()
    if ((align % 4)>0): 
        #skip padding to next 4-byte boundary
        stream.read(4 - (align % 4))

class decoder(object):
    _struct = ''
    _fields = {}
    _flags = {}
    
    def __init__(self, stream):
        self._value = struct.unpack(self._struct,
            stream.read(struct.calcsize(self._struct)))
    
    def __getattr__(self, item):
        if item in self._fields.keys():
            keyid = self._fields[item]
            return self._value[keyid]
        elif item in self._flags.keys():
            flag, bit = self._flags[item]
            return bool(self._value[flag] & bit)
        else:
            raise AttributeError(item)

    def __repr__(self):
        return repr(self._value)
    
#***********************************************
# class definitions for anm file
#***********************************************   

class anm_header(decoder):
    _struct = "LBBHffffffffffffBxxx"
    _fields = {
        "Magic":0,
        "count_anims":1,
        "count_static_anims":2,
        "count_frames":3,
        "m11":4,
        "m12":5,
        "m13":6,
        "m21":7,
        "m22":8,
        "m23":9,
        "m31":10,
        "m32":11,
        "m33":12,
        "offset_x":13,
        "offset_y":14,
        "offset_z":15,
        } 
    _flags = {
        "is_variance":(16,1),
        "is_30fps": (16,2)
        }
    def __init__(self, stream):
        global origins
        decoder.__init__(self, stream)
        m1 = [self.m11, self.m12, self.m13]
        m2 = [self.m21, self.m22, self.m23]
        m3 = [self.m31, self.m32, self.m33]
        m4 = [self.offset_x, self.offset_y, self.offset_z]
        self.matrix = Matrix(m1, m2, m3, m4).resize4x4()
        origins += [self.matrix]
class anm_origin(decoder):
    _struct = "ffffffffffff"
    _fields = {
        "m11":0,
        "m12":1,
        "m13":2,
        "m21":3,
        "m22":4,
        "m23":5,
        "m31":6,
        "m32":7,
        "m33":8,
        "offset_x":9,
        "offset_y":10,
        "offset_z":11,
        } 
    def __init__(self, stream):
        global origins
        decoder.__init__(self, stream)
        m1 = [self.m11, self.m12, self.m13]
        m2 = [self.m21, self.m22, self.m23]
        m3 = [self.m31, self.m32, self.m33]
        m4 = [self.offset_x, self.offset_y, self.offset_z]
        self.origin = Matrix(m1, m2, m3, m4).resize4x4()
        origins += [self.origin]        
class bone_anm_header(decoder):
    _struct = "LHBBffff"
    _fields = {
        "frame_count": 0,
        "bone_count": 1,
        "besson_type": 2, 
        "dest_x": 4,
        "dest_y": 5,
        "dest_z": 6, 
        "dest_angle": 7
        }
    _flags = {
        "has_scale": (3,1),
        "has_probability": (3,2),
        "has_animated_scale": (3,4)
        }
    def __init__(self, stream):
        global compression
        decoder.__init__(self, stream)
        if self.has_probability:
            self.probability = read_item("f", stream)
        if self.has_scale:
            self.compression = read_item("f", stream)
#            print 'compression = ',self.compression
            compression += [self.compression]
    
class BoneFrames(list):
    def __init__(self, stream, framecount):
        #bf = [None] * framecount
        bf = []
        mask = read_item("H", stream)
        fc = framecount
        while fc > 0:
            holdcount = 1
            data = read_item("H", stream)
            if (data & mask):
                v1 = data & (mask - 1)
                v2 = data - mask
                v3 = ~(mask - 1)
                holdcount = v1 | ((v2 & v3) >> 1)
                if (holdcount>framecount) or (holdcount<0):
                    raise ValueError, 'value is outside bounds for hold.'
                holdvalue = UncompressU16(read_item("H", stream))
                for hold in range(holdcount):
                    bf.append(holdvalue)
            else:
                bf.append(UncompressU16(data))
            fc -= holdcount
        super(BoneFrames, self).__init__(bf)
        
class BoneAnim(object):
    def __init__(self, stream, framecount):
        self.trans_static = read_item("bxxx", stream)
        self.x = BoneFrames(stream, framecount)
        self.y = BoneFrames(stream, framecount)
        self.z = BoneFrames(stream, framecount)
        pad(stream)
        if self.trans_static == 0:
            self.u = BoneFrames(stream, framecount)
            self.v = BoneFrames(stream, framecount)
            self.w = BoneFrames(stream, framecount)
        else:
            self.u = [read_item("f", stream)] * framecount
            self.v = [read_item("f", stream)] * framecount
            self.w = [read_item("f", stream)] * framecount
        pad(stream)
    
class BoneAnims(list):
    def __init__(self, stream):
        self.header = bone_anm_header(stream)        
        self.bone_list = []
        for b in range(self.header.bone_count):
            self.bone_list.append(read_item("b", stream))
        pad(stream)
        ba_list = []
        for ba in range(self.header.bone_count):
            ba_list.append(BoneAnim(stream, self.header.frame_count))
#            print 'ba_list ',bonelist.bns[ba],' trans_static ',ba_list[ba].trans_static
        
        BoneAnms = zip(self.bone_list, ba_list)
        super(BoneAnims, self).__init__(BoneAnms)

                
class prop_anm_header(object):
    def __init__(self, stream):
        val = read_item("Hxx", stream)
        self.is_carriable = val & 1
        self.has_scale = (val & 2)>>1
        self.index_like_me = ((val | 32707) - 32707) >> 2 #This is 4 bits wide.. 0-15 valid values
        self.is_more_than_one_like_me = (val & 64) >> 6
        self.index_skeleton = (val | 127) >> 7 #This is the remainder of 8 bits wide...0-255 valid values
        
        if self.has_scale:
            self.compression = read_item("f", stream)

class PropFrame(object):
    def __init__(self, stream):
        self.x = UncompressU16(read_item("H",stream))
        self.y = UncompressU16(read_item("H",stream))
        self.z = UncompressU16(read_item("H",stream))
        self.u = UncompressU16(read_item("H",stream))
        self.v = UncompressU16(read_item("H",stream))
        self.w = UncompressU16(read_item("H",stream))
             
class PropFrames(list):
    def __init__(self, stream, framecount):
        PropPos = [None] * framecount
        for j in range(framecount):
            PropPos[j] = PropFrame(stream)
        super(PropFrames, self).__init__(PropPos)

    
class PropAnim(object):
    def __init__(self, stream, framecount):
        self.name = read_nts(stream)
        self.header = prop_anm_header(stream)
        self.frames = PropFrames(stream, framecount)
    
def quat_mult(r,q):
    p0 = r[0]*q[0]-r[1]*q[1]-r[2]*q[2]-r[3]*q[3]
    p1 = r[0]*q[1]+r[1]*q[0]-r[2]*q[3]+r[3]*q[2]
    p2 = r[0]*q[2]+r[1]*q[3]+r[2]*q[0]-r[3]*q[1]
    p3 = r[0]*q[3]-r[1]*q[2]+r[2]*q[1]+r[3]*q[0]
    p = Blender.Mathutils.Quaternion(p0,p1,p2,p3)
    return p

def set_bqr(x,y,z):
    e = Blender.Mathutils.Euler(x,y,z)
    q = e.toQuat()
    return q[:]
#***********************************************
def tm_to_blender(x,y,z):
	ea = Blender.Mathutils.Euler(0.0,0.0,y)
	mz = ea.toMatrix()
	ea = Blender.Mathutils.Euler(z,0.0,0.0)
	mx = ea.toMatrix()
	ea = Blender.Mathutils.Euler(0.0,x,0.0)
	my = ea.toMatrix()
	m = my*mz*mx
	quat = m.toQuat()
	return quat
def get_armatures():
    obs = []
    for ob in scn.objects:
        print 'ob.type = ',ob.type,' name = ',ob.name
        if ob.type == 'Armature':
#		    arm_ob = ob
#                    print 'ob.name = ',ob.name
		    obs += [ob]
    return obs
def get_empty(name):
    print 'looking for ',name
#    if scn.properties.has_key('Static Props') == False:
#        scn.properties['Static Props'] = {}
#    scn.properties['Static Props'][name] = 0
    for ob in scn.objects:
        print 'ob.type = ',ob.type,' name = ',ob.name
        if ob.type == 'Empty':
            if ob.properties.has_key('TheMovies') == True:
                if ob.properties['TheMovies'].has_key('grpName'):
                    print 'Object name = ',ob.name
                    print 'group name = ',name
                    grpName = ob.properties['TheMovies']['grpName']
                    if grpName == name:
                        print 'found match ',ob.name
#                    scn.properties['Static Props'][name] = 1
                        return ob
    return scn.objects.new('Empty',name)
#    return 0
def clear_animation(arm_ob,pbones,f0,framecount):
    for pbone in pbones:
        scn = Blender.Scene.GetCurrent()
        

        for f in range(framecount):
            pbone.quat = Quaternion(1.,0,0,0)
            pbone.loc = Vector(0,0,0)            
            pbone.insertKey(arm_ob,f+f0,Blender.Object.Pose.ROT)
            pbone.insertKey(arm_ob,f+f0,Blender.Object.Pose.LOC)
#        scn.update()
#    time.sleep(5.0)
def load_anm (filename):
    start_time = Blender.sys.time()
    prev_time = start_time
    in_editmode = Window.EditMode()
    if in_editmode:
        Window.EditMode(0)
    stream = open(filename,"rb")
    filepath = os.path.dirname(filename)
    filebase = os.path.basename(filename)
    fullbase = os.path.splitext(filebase)[0]
    basename = fullbase[0:21] #Blender has a $%$@ 21-char limit!
    basedir = os.path.split(filepath)[0]
    #costume_bone_name_list()
#    horse_bone_name_list()
    scene = Blender.Scene.GetCurrent()
    #scene.setLayers([1])

    #Start Reading!
    header = anm_header(stream)
    if header.Magic != 12:
        raise IOError, "File is not a TM Animation"
    #Get origin matrices for each subsequent animation -- first must use OriginInWorld?
    print 'header processed: %.4f sec.' % (Blender.sys.time()-prev_time)
    prev_time = Blender.sys.time()
    # set ID properties for anm_header
    
    framecount = header.count_frames
    animorigins = []
    animorigins.append(header.matrix)
    for a in range(header.count_anims-1):
        animorigins.append(anm_origin(stream).origin)
    #Read bone-based anim data
    bone_anims = {}
    for ba in range(header.count_anims):
        bone_anims[ba] = BoneAnims(stream)
    #Read prop anim data
    print '%d Bone Anims processed: %.4f sec.' % (header.count_anims, (Blender.sys.time()-prev_time))
    prev_time = Blender.sys.time()
    prop_anims = {}
    for pa in range(header.count_static_anims):
        tprop = PropAnim(stream, framecount)
        prop_anims[tprop.name] = tprop
        print 'tprop.name = ',tprop.name
    print '%d Prop Anims processed: %.4f sec.' % (header.count_static_anims, (Blender.sys.time()-prev_time))
    prev_time = Blender.sys.time()
    print '%s: %d Frames Imported: %.4f sec.' % (basename, framecount, (Blender.sys.time() - start_time))
    #Now do something with it! Translate into Blender stuff
    #Check for armatures
    arms = Blender.Armature.Get()
    if len(arms) == 0:
        block = []
        block.append("No armature found.")
        block.append("Importing points only.")
        retval = Blender.Draw.PupBlock("WARNING", block)
    if len(arms) >0 and len(arms) < len(animorigins):
        block = []
        block.append("Too few armatures found.")
        block.append("Anim needs %d armatures." % header.count_anims)
        block.append("Some will be points only.")
        retval = Blender.Draw.PupBlock("WARNING", block)   
    if len(arms) >0 and len(arms) > len(animorigins):
        block = []
        block.append("Too many armatures found.")
        block.append("Anim needs %d armatures." % header.count_anims)
        block.append("Some will be ignored.")
        retval = Blender.Draw.PupBlock("WARNING", block) 



#    scn = Blender.Scene.GetCurrent()
    if scn.properties.has_key('ANMHead') == False:
        scn.properties['ANMHead'] = {}
    scn.properties['ANMHead']['Target File'] = filename        
    scn.properties['ANMHead']['count_anims'] = header.count_anims
    scn.properties['ANMHead']['count_static'] = header.count_static_anims
    for i in range(len(animorigins)):
        name = 'Origin'+str(i)
        (roll,pitch,yaw) = (0.0,0.0,0.0)
        origin = animorigins[i]
#        print origin
        (roll,pitch,yaw) = origin.rotationPart().toEuler()
        (x_origin,y_origin,z_origin) = origin.translationPart()
        if scn.properties['ANMHead'].has_key(name) == False:
            scn.properties['ANMHead'][name] = {}
        scn.properties['ANMHead'][name]['yaw'] = yaw
        scn.properties['ANMHead'][name]['pitch'] = pitch
        scn.properties['ANMHead'][name]['roll'] = roll
        scn.properties['ANMHead'][name]['x_origin'] = x_origin
        scn.properties['ANMHead'][name]['y_origin'] = y_origin
        scn.properties['ANMHead'][name]['z_origin'] = z_origin
    for i in range(header.count_anims):
        name = 'BoneAnmHdr' + str(i+1)
        if scn.properties.has_key(name) == False:
            scn.properties[name] = {}
        scn.properties[name]['compression'] = bone_anims[i].header.compression
        scn.properties[name]['dest_x'] = bone_anims[i].header.dest_x
        scn.properties[name]['dest_y'] = bone_anims[i].header.dest_y
        scn.properties[name]['dest_z'] = bone_anims[i].header.dest_z
        scn.properties[name]['dest_angle'] = bone_anims[i].header.dest_angle
        bone_anim = bone_anims[i]
        bonelist.bone_name_list(bone_anim.header.bone_count)
        bns = bonelist.bns
# Mark all bones with non-static position data
        for btag in range(len(bone_anim)):         
            if bone_anim[btag][1].trans_static == 0:
                scn.properties[name][bns[btag]] = 'non-static'    
    arms = Blender.Armature.Get()
    obs = get_armatures()
    nob = -1
    if scn.properties.has_key('append') == False:
        scn.properties['append'] = 0
    if scn.properties['append'] == 1:
        f0 = Blender.Get('endframe')
    else:
        f0 = Blender.Get('staframe')
    for ba in bone_anims.keys():
        nob += 1
        arm_ob = obs[nob]
        arm_data = arm_ob.getData()
        arm_data.makeEditable()
        bones = arm_data.bones.values()
        pose = arm_ob.getPose()
        pbones = pose.bones.values()
        pbdict = {}
        for pbone in pbones:
            pbdict[pbone.name] = pbone 
        acts = Blender.Armature.NLA.GetActions()

        if len(acts) < len(obs):
            act = Blender.Armature.NLA.NewAction()
        else:
            key = acts.keys()[nob]
            act = acts[key]
        act.setActive(arm_ob)
        btag = 0
        xoff = origins[ba][3][0]
        yoff = origins[ba][3][1]
        zoff = origins[ba][3][2]
        rot = origins[ba].rotationPart()
        kf_non_trans_static = compression[ba]
        print ' '
        print 'processing bone anim ',ba
#        pdb.set_trace()
        nbones = bone_anims[ba].header.bone_count
        print 'nbones  = ',nbones
        bonelist.bone_name_list(nbones)
        bns = bonelist.bns
        root_bone = bns[0]
        print 'ba = %d root_bone = %s',ba,root_bone
        locs = bonelist.locs
        bqr = bonelist.bqr
#        clear_animation(arm_ob,pbones,f0,framecount)
#        pose.update()
        btag = 0
        for bi, ba_obj in bone_anims[ba]:
                    print btag,bi,bns[btag]
                    pbone = pbdict[bns[btag]]
                    rs = 180.0
                    bs = bqr[btag]
                    bq = Blender.Mathutils.Quaternion(bs[0],bs[1],bs[2],bs[3]) 
                    kf = 1.0
                    if ba_obj.trans_static == 0:
                        kf = kf_non_trans_static
#                    parent = pbone.parent
#                    if parent == None:
#                        PP = Matrix([0,0,1],[1,0,0],[0,1,0])
#                    else :
#                        PP = parent.poseMatrix.rotationPart()
#                    R = pbone.quat.toMatrix()
#                    P = pbone.poseMatrix.rotationPart().transpose()
#                    RP = PP*P*R
#                    bq = RP.toQuat()
                    print 'bq = ',bq
                    for f in range(framecount):
                        if f == 0 and btag == 0:
                            print 'first frame root'
                            print pbone.quat
                            print pbone.loc
                        u = rs*ba_obj.x[f]
                        v = rs*ba_obj.y[f]
                        w = rs*ba_obj.z[f]
                        x = ba_obj.u[f]*kf-locs[btag][0]
                        y = ba_obj.v[f]*kf-locs[btag][1]
                        z = ba_obj.w[f]*kf-locs[btag][2]
                        pbone.quat = Quaternion(1,0,0,0)
                        pbone.loc = Vector(0,0,0)
                        pose.update()
                        quat = tm_to_blender(u,v,w)
                        pbone.quat = quat_mult(quat,bq)
                        pm = pbone.poseMatrix.rotationPart()
                        pbone.insertKey(arm_ob,f+f0,Blender.Object.Pose.ROT)
                        rel = Vector(0,0,0)
                        if btag == 0:
                            rel = pm*Vector(x,y,z)
                        pbone.loc[:] = rel
                        pbone.insertKey(arm_ob,f+f0,Blender.Object.Pose.LOC)
                        pose.update()
                    btag += 1
        pose.update()
# now prop anim.
    print 'prop_anims.keys = ',prop_anims.keys()
    print ' f0 = ',f0
    for pa in prop_anims.keys():
        prop_anim = prop_anims[pa]
        prop_ob = get_empty(prop_anim.name)
        if prop_ob:
            print pa,prop_anim.name,prop_ob.name
            comp = prop_anim.header.compression
            print 'compression = ',prop_anim.header.compression
            print 'Frame Data ',f0
            for f in range(framecount):
                Blender.Set('curframe',f+f0)
                frame = prop_anim.frames[f]
                (u,v,w) = (math.pi*frame.u,math.pi*frame.v,math.pi*frame.w)
                (x,y,z) = (comp*frame.x,comp*frame.y,comp*frame.z)
                prop_ob.setEuler(u,v,w)
                prop_ob.setLocation(x,y,z)
                prop_ob.insertIpoKey(Blender.Object.ROT)
                prop_ob.insertIpoKey(Blender.Object.LOC)
            
            

    print 'basedir = ',basedir  
    print 'filename = ',filename          
    context = scene.getRenderingContext()
    context.startFrame(1)
    context.endFrame(framecount+f0-1)
    Blender.Scene.GetCurrent().update()

#***********************************************
# register callback
#***********************************************
def my_callback(filename):
    load_anm(filename)

Blender.Window.FileSelector(my_callback, "Import ANM", '*.anm')
#load_anm('c:/TMBlender/data/animations/high/tired_breathe_idle.anm')
